#include "primitive.hpp"
#include <Eigen/Dense>

double random_scalar() {
	return rand() / (double)RAND_MAX * 2.0 - 1.0;
}

Eigen::Vector3d random_vec() {
	return Eigen::Vector3d(random_scalar(), random_scalar(), random_scalar());
}

Eigen::Vector3d random_dir() {
	Eigen::Vector3d n = random_vec();
	return n / n.norm();
}

bool same_space(const Eigen::MatrixXd& A, const Eigen::MatrixXd& B) {
	if (A.cols() != B.cols() || A.rows() != 6)
		return false;

	{
		Eigen::JacobiSVD<Eigen::MatrixXd> svd(A, Eigen::ComputeThinU | Eigen::ComputeThinV);
		Eigen::VectorXd sig = svd.singularValues();
		if (sig[A.cols() - 1] < 1e-6)
			return false;
	}
	Eigen::MatrixXd C(A.rows(), A.cols()+B.cols());
	C << A, B;
	Eigen::JacobiSVD<Eigen::MatrixXd> svd(C, Eigen::ComputeThinU | Eigen::ComputeThinV);
	Eigen::VectorXd sig = svd.singularValues();

	return sig[B.cols()] / sig[0] < 1e-6;
}


bool test_plane()
{
	for (int test_examples = 0; test_examples < 10; ++test_examples) {
		// simulate the point cloud
		Eigen::MatrixXd V(128, 3);
		Eigen::MatrixXd N(128, 3);
		Signature signature;
		Eigen::MatrixXd gt_space(6, 3);

		Eigen::Vector3d n = random_vec();
		Eigen::Vector3d tx = random_vec();
		Eigen::Vector3d ty = n.cross(tx);
		Eigen::Vector3d c = random_vec();
		tx = ty.cross(n);
		tx /= tx.norm();
		ty /= ty.norm();

		std::set<int> s;
		for (int i = 0; i < 128; ++i) {
			V.row(i) = tx * random_scalar() + ty * random_scalar() + c;
			N.row(i) = n;
			s.insert(i);
		}
		EstimatePrimitive(V, N, s, signature);
		gt_space.col(0).head(3) = n;
		gt_space.col(0).tail(3) = Eigen::Vector3d::Zero();
		gt_space.col(1).head(3) = Eigen::Vector3d::Zero();
		gt_space.col(1).tail(3) = tx;
		gt_space.col(2).head(3) = Eigen::Vector3d::Zero();
		gt_space.col(2).tail(3) = ty;

		if (!same_space(signature.null_space, gt_space))
			return false;
	}

	return true;
}

bool test_sphere()
{
	for (int test_examples = 0; test_examples < 10; ++test_examples) {
		Eigen::MatrixXd V(128, 3);
		Eigen::MatrixXd N(128, 3);
		Signature signature;

		double r = random_scalar();
		Eigen::Vector3d c = random_vec();
		std::set<int> s;
		for (int i = 0; i < 128; ++i) {
			Eigen::Vector3d d = random_dir();
			V.row(i) = c + r * d;
			N.row(i) = d;
			s.insert(i);
		}

		EstimatePrimitive(V, N, s, signature);

		Eigen::MatrixXd gt_space = Eigen::MatrixXd::Zero(6, 3);

		Eigen::Vector3d t1 = Eigen::Vector3d(1, 0, 0).cross(c);
		Eigen::Vector3d t2 = Eigen::Vector3d(0, 1, 0).cross(c);
		Eigen::Vector3d t3 = Eigen::Vector3d(0, 0, 1).cross(c);
		gt_space(0, 0) = 1;
		gt_space.col(0).tail(3) = -t1;
		gt_space(1, 1) = 1;
		gt_space.col(1).tail(3) = -t2;
		gt_space(2, 2) = 1;
		gt_space.col(2).tail(3) = -t3;

		if (!same_space(signature.null_space, gt_space))
			return false;

	}
	return true;
}

bool test_cylinder()
{
	for (int test_examples = 0; test_examples < 10; ++test_examples) {
		Eigen::MatrixXd V(128, 3);
		Eigen::MatrixXd N(128, 3);
		Signature signature;

		double r = random_scalar();
		Eigen::Vector3d c = random_vec();
		Eigen::Vector3d n = random_vec();
		Eigen::Vector3d tx = random_vec();
		Eigen::Vector3d ty = n.cross(tx);

		tx = ty.cross(n);
		tx /= tx.norm();
		ty /= ty.norm();

		std::set<int> s;
		for (int i = 0; i < 128; ++i) {
			double theta = random_scalar() * 2 * 3.141592654;
			V.row(i) = c + r * (tx * cos(theta) + ty * sin(theta)) + n * random_scalar();
			N.row(i) = tx * cos(theta) + ty * sin(theta);
			s.insert(i);
		}

		EstimatePrimitive(V, N, s, signature);

		Eigen::MatrixXd gt_space = Eigen::MatrixXd::Zero(6, 2);
		gt_space.col(0).tail(3) = n;
		gt_space.col(1).head(3) = n;
		gt_space.col(1).tail(3) = -n.cross(c);

		if (!same_space(signature.null_space, gt_space))
			return false;

	}
	return true;	
}

bool test_cone()
{
	for (int test_examples = 0; test_examples < 10; ++test_examples) {
		Eigen::MatrixXd V(128, 3);
		Eigen::MatrixXd N(128, 3);
		Signature signature;

		double alpha = std::abs(random_scalar()) * 3.141592654;
		Eigen::Vector3d c = random_vec();
		Eigen::Vector3d n = random_vec();
		Eigen::Vector3d tx = random_vec();
		Eigen::Vector3d ty = n.cross(tx);

		tx = ty.cross(n);
		tx /= tx.norm();
		ty /= ty.norm();

		std::set<int> s;
		for (int i = 0; i < 128; ++i) {
			double theta = random_scalar() * 2 * 3.141592654;
			double dis = random_scalar();
			double r = dis * tan(alpha);
			Eigen::Vector3d offset = r * (tx * cos(theta) + ty * sin(theta)) + n * dis;
			Eigen::Vector3d sx = offset.cross(n);
			Eigen::Vector3d sn = offset.cross(sx);
			V.row(i) = c + offset;
			N.row(i) = sn / sn.norm();
			s.insert(i);
		}

		EstimatePrimitive(V, N, s, signature);

		Eigen::MatrixXd gt_space = Eigen::MatrixXd::Zero(6, 1);
		gt_space.col(0).head(3) = n;
		gt_space.col(0).tail(3) = -n.cross(c);

		if (!same_space(signature.null_space, gt_space))
			return false;

	}
	return true;	
}

int main(int argc, char** argv) {
	
	if (test_plane()) {
		printf("Test Plane Success!\n");
	} else {
		printf("Test Plane Failed!\n");
	}

	if (test_sphere()) {
		printf("Test Sphere Success!\n");
	} else {
		printf("Test Sphere Failed!\n");
	}

	if (test_cylinder()) {
		printf("Test Cylinder Success!\n");
	} else {
		printf("Test Cylinder Failed!\n");
	}

	if (test_cone()) {
		printf("Test Cone Success!\n");
	} else {
		printf("Test Cone Failed!\n");
	}

	return 0;
}