import random
import math
import numpy as np
import tensorflow as tf
DIVISION_EPS = 1e-10
SQRT_EPS = 1e-10
LS_L2_REGULARIZER = 1e-8

def sqrt_safe(x):
    return tf.sqrt(tf.abs(x) + SQRT_EPS)

def weighted_plane_fitting(P, W):
    # P - BxNx3
    # W - BxN
    # Returns n, c, with n - Bx3, c - B
    WP = P * tf.expand_dims(W, axis=2) # BxNx3
    W_sum = tf.reduce_sum(W, axis=1) # B
    P_weighted_mean = tf.reduce_sum(WP, axis=1) / tf.maximum(tf.expand_dims(W_sum, 1), DIVISION_EPS) # Bx3
    A = P - tf.expand_dims(P_weighted_mean, axis=1) # BxNx3
    n = solve_weighted_tls(A, W) # Bx3
    c = tf.reduce_sum(n * P_weighted_mean, axis=1)
    return n, c

def custom_svd_v_column(M, col_index=-1):
    # Must make sure M is finite. Otherwise cudaSolver might fail.
    assert_op = tf.Assert(tf.logical_not(tf.reduce_any(tf.logical_not(tf.is_finite(M)))), [M], summarize=10)
    with tf.control_dependencies([assert_op]):
        with tf.get_default_graph().gradient_override_map({'Svd': 'CustomSvd'}):
            s, u, v = tf.svd(M, name='Svd') # M = usv^T
            return v[:, :, col_index] 

def guarded_matrix_solve_ls(A, b, W, condition_number_cap=1e5):
    # Solve weighted least square ||\sqrt(W)(Ax-b)||^2
    # A - BxNxD
    # b - BxNx1
    # W - BxN

    sqrt_W = tf.sqrt(tf.maximum(W, SQRT_EPS)) # BxN
    A *= tf.expand_dims(sqrt_W, axis=2) # BxNxD
    b *= tf.expand_dims(sqrt_W, axis=2) # BxNx1
    # Compute singular value, trivializing the problem when condition number is too large
    AtA = tf.matmul(a=A, b=A, transpose_a=True)
    s, _, _ = [tf.stop_gradient(u) for u in tf.svd(AtA)] # s will be BxD
    mask = tf.less(s[:, 0] / s[:, -1], condition_number_cap) # B
    A *= tf.to_float(tf.expand_dims(tf.expand_dims(mask, axis=1), axis=2)) # zero out badly conditioned data
    x = tf.matrix_solve_ls(A, b, l2_regularizer=LS_L2_REGULARIZER, fast=True) # BxDx1 
    return tf.squeeze(x, axis=2) # BxD

def solve_weighted_tls(A, W):
    # A - BxNx3
    # W - BxN, positive weights
    # Find solution to min x^T A^T diag(W) A x = min ||\sqrt{diag(W)} A x||^2, subject to ||x|| = 1
    A_p = tf.expand_dims(A, axis=2) * tf.expand_dims(A, axis=3) # BxNx3x3
    W_p = tf.expand_dims(tf.expand_dims(W, axis=2), axis=3) # BxNx1x1
    M = tf.reduce_sum(W_p * A_p, axis=1) # Bx3x3
    x = custom_svd_v_column(M) # Bx3
    return x

def weighted_sphere_fitting(P, W):
    # P - BxNxD
    # W - BxN
    W_sum = tf.reduce_sum(W, axis=1) # B
    WP_sqr_sum = tf.reduce_sum(W * tf.reduce_sum(tf.square(P), axis=2), axis=1) # B
    P_sqr = tf.reduce_sum(tf.square(P), axis=2) # BxN
    b = tf.expand_dims(tf.expand_dims(WP_sqr_sum / tf.maximum(W_sum, DIVISION_EPS), axis=1) - P_sqr, axis=2) # BxNx1
    WP_sum = tf.reduce_sum(tf.expand_dims(W, axis=2) * P, axis=1) # BxD
    A = 2 * (tf.expand_dims(WP_sum / tf.expand_dims(tf.maximum(W_sum, DIVISION_EPS), axis=1), axis=1) - P) # BxNxD

    # Seek least norm solution to the least square
    center = guarded_matrix_solve_ls(A, b, W) # BxD
    W_P_minus_C_sqr_sum = P - tf.expand_dims(center, axis=1) # BxNxD
    W_P_minus_C_sqr_sum = W * tf.reduce_sum(tf.square(W_P_minus_C_sqr_sum), axis=2) # BxN
    r_sqr = tf.reduce_sum(W_P_minus_C_sqr_sum, axis=1) / tf.maximum(W_sum, DIVISION_EPS) # B

    return {'center': center, 'radius_squared': r_sqr}

def batched_gather(data, indices, axis):
    # data - Bx...xKx..., axis is where dimension K is
    # indices - BxK
    # output[b, ..., k, ...] = in[b, ..., indices[b, k], ...]
    assert axis >= 1
    ndims = data.get_shape().ndims # allow dynamic rank
    if axis > 1:
        # tranpose data to BxKx...
        perm = np.arange(ndims)
        perm[axis] = 1
        perm[1] = axis
        data = tf.transpose(data, perm=perm)
    batch_size = tf.shape(data)[0]
    batch_nums = tf.tile(tf.expand_dims(tf.expand_dims(tf.range(batch_size), axis=1), axis=2), multiples=[1, tf.shape(indices)[1], 1]) # BxKx1
    indices = tf.concat([batch_nums, tf.expand_dims(indices, axis=2)], axis=2) # BxKx2
    gathered_data = tf.gather_nd(data, indices=indices)
    if axis > 1:
        gathered_data = tf.transpose(gathered_data, perm=perm)
    return gathered_data

def adaptor_matching(param_list, matching_indices):
    return [tf.expand_dims(batched_gather(param, matching_indices, axis=1), axis=2) for param in param_list]
